package com.sun.java.util.jar.pack;

import java.util.AbstractList;
import java.util.ListIterator;
import java.util.Set;
import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Map;

abstract class ConstantPool
{
    protected static final ConstantPool.Entry[] noRefs;
    protected static final ConstantPool.ClassEntry[] noClassRefs;
    static final byte[] TAGS_IN_ORDER;
    static final byte[] TAG_ORDER;
    static final byte[] NUMBER_TAGS;
    static final byte[] EXTRA_TAGS;
    static final byte[] LOADABLE_VALUE_TAGS;
    static final byte[] ANY_MEMBER_TAGS;
    static final byte[] FIELD_SPECIFIC_TAGS;
    
    private ConstantPool() {
    }
    
    static int verbose() {
        return Utils.currentPropMap().getInteger("com.sun.java.util.jar.pack.verbose");
    }
    
    public static synchronized ConstantPool.Utf8Entry getUtf8Entry(String s) {
        Map<String, ConstantPool.Utf8Entry> utf8Entries = Utils.getTLGlobals().getUtf8Entries();
        ConstantPool.Utf8Entry utf8Entry = (ConstantPool.Utf8Entry)utf8Entries.get(s);
        if (utf8Entry == null) {
            utf8Entry = new Utf8Entry(s);
            utf8Entries.put(utf8Entry.stringValue(), utf8Entry);
            return utf8Entry;
        }
        return utf8Entry;
    }
    
    public static ConstantPool.ClassEntry getClassEntry(String s) {
        Map<String, ConstantPool.ClassEntry> classEntries = Utils.getTLGlobals().getClassEntries();
        ConstantPool.ClassEntry classEntry = (ConstantPool.ClassEntry)classEntries.get(s);
        if (classEntry == null) {
            classEntry = new ConstantPool.ClassEntry(getUtf8Entry(s));
            assert !(!s.equals(classEntry.stringValue()));
            classEntries.put(classEntry.stringValue(), classEntry);
        }
        return classEntry;
    }
    
    public static ConstantPool.LiteralEntry getLiteralEntry(Comparable<?> comparable) {
        Map<Comparable<?>, ConstantPool.LiteralEntry> literalEntries = Utils.getTLGlobals().getLiteralEntries();
        ConstantPool.LiteralEntry literalEntry = (ConstantPool.LiteralEntry)literalEntries.get(comparable);
        if (literalEntry == null) {
            if ((comparable instanceof String)) {
                literalEntry = new ConstantPool.StringEntry(getUtf8Entry((String)comparable));
            }
            else {
                literalEntry = new ConstantPool.NumberEntry((Number)comparable);
            }
            literalEntries.put(comparable, literalEntry);
            return literalEntry;
        }
        else {
            return literalEntry;
        }
    }
    
    public static ConstantPool.StringEntry getStringEntry(String s) {
        return (ConstantPool.StringEntry)getLiteralEntry(s);
    }
    
    public static ConstantPool.SignatureEntry getSignatureEntry(String anObject) {
        Map<String, ConstantPool.SignatureEntry> signatureEntries = Utils.getTLGlobals().getSignatureEntries();
        ConstantPool.SignatureEntry signatureEntry = (ConstantPool.SignatureEntry)signatureEntries.get(anObject);
        if (signatureEntry == null) {
            signatureEntry = new SignatureEntry(anObject);
            assert !(!signatureEntry.stringValue().equals(anObject));
            signatureEntries.put(anObject, signatureEntry);
        }
        return signatureEntry;
    }
    
    public static ConstantPool.SignatureEntry getSignatureEntry(ConstantPool.Utf8Entry utf8Entry, ConstantPool.ClassEntry[] array) {
        return getSignatureEntry(ConstantPool.SignatureEntry.stringValueOf(utf8Entry, array));
    }
    
    public static ConstantPool.DescriptorEntry getDescriptorEntry(ConstantPool.Utf8Entry utf8Entry, ConstantPool.SignatureEntry signatureEntry) {
        Map<String, ConstantPool.DescriptorEntry> descriptorEntries = Utils.getTLGlobals().getDescriptorEntries();
        String stringValue = ConstantPool.DescriptorEntry.stringValueOf(utf8Entry, signatureEntry);
        ConstantPool.DescriptorEntry descriptorEntry = (ConstantPool.DescriptorEntry)descriptorEntries.get(stringValue);
        if (descriptorEntry == null) {
            descriptorEntry = new ConstantPool.DescriptorEntry(utf8Entry, signatureEntry);
            assert !(!descriptorEntry.stringValue().equals(stringValue)) : descriptorEntry.stringValue() + " != " + stringValue;
            descriptorEntries.put(stringValue, descriptorEntry);
        }
        return descriptorEntry;
    }
    
    public static ConstantPool.DescriptorEntry getDescriptorEntry(ConstantPool.Utf8Entry utf8Entry, ConstantPool.Utf8Entry utf8Entry2) {
        return getDescriptorEntry(utf8Entry, getSignatureEntry(utf8Entry2.stringValue()));
    }
    
    public static ConstantPool.MemberEntry getMemberEntry(byte b, ConstantPool.ClassEntry classEntry, ConstantPool.DescriptorEntry descriptorEntry) {
        Map<String, ConstantPool.MemberEntry> memberEntries = Utils.getTLGlobals().getMemberEntries();
        String stringValue = ConstantPool.MemberEntry.stringValueOf(b, classEntry, descriptorEntry);
        ConstantPool.MemberEntry memberEntry = (ConstantPool.MemberEntry)memberEntries.get(stringValue);
        if (memberEntry == null) {
            memberEntry = new ConstantPool.MemberEntry(b, classEntry, descriptorEntry);
            assert !(!memberEntry.stringValue().equals(stringValue)) : memberEntry.stringValue() + " != " + stringValue;
            memberEntries.put(stringValue, memberEntry);
        }
        return memberEntry;
    }
    
    public static ConstantPool.MethodHandleEntry getMethodHandleEntry(byte b, ConstantPool.MemberEntry memberEntry) {
        Map<String, ConstantPool.MethodHandleEntry> methodHandleEntries = Utils.getTLGlobals().getMethodHandleEntries();
        String stringValue = ConstantPool.MethodHandleEntry.stringValueOf((int)b, memberEntry);
        ConstantPool.MethodHandleEntry methodHandleEntry = (ConstantPool.MethodHandleEntry)methodHandleEntries.get(stringValue);
        if (methodHandleEntry == null) {
            methodHandleEntry = new ConstantPool.MethodHandleEntry(b, memberEntry);
            assert !(!methodHandleEntry.stringValue().equals(stringValue));
            methodHandleEntries.put(stringValue, methodHandleEntry);
        }
        return methodHandleEntry;
    }
    
    public static ConstantPool.MethodTypeEntry getMethodTypeEntry(ConstantPool.SignatureEntry signatureEntry) {
        Map<String, ConstantPool.MethodTypeEntry> methodTypeEntries = Utils.getTLGlobals().getMethodTypeEntries();
        String stringValue = signatureEntry.stringValue();
        ConstantPool.MethodTypeEntry methodTypeEntry = (ConstantPool.MethodTypeEntry)methodTypeEntries.get(stringValue);
        if (methodTypeEntry == null) {
            methodTypeEntry = new MethodTypeEntry(signatureEntry);
            assert !(!methodTypeEntry.stringValue().equals(stringValue));
            methodTypeEntries.put(stringValue, methodTypeEntry);
        }
        return methodTypeEntry;
    }
    
    public static ConstantPool.MethodTypeEntry getMethodTypeEntry(ConstantPool.Utf8Entry utf8Entry) {
        return getMethodTypeEntry(getSignatureEntry(utf8Entry.stringValue()));
    }
    
    public static ConstantPool.InvokeDynamicEntry getInvokeDynamicEntry(ConstantPool.BootstrapMethodEntry bootstrapMethodEntry, ConstantPool.DescriptorEntry descriptorEntry) {
        Map<String, ConstantPool.InvokeDynamicEntry> invokeDynamicEntries = Utils.getTLGlobals().getInvokeDynamicEntries();
        String stringValue = ConstantPool.InvokeDynamicEntry.stringValueOf(bootstrapMethodEntry, descriptorEntry);
        ConstantPool.InvokeDynamicEntry invokeDynamicEntry = (ConstantPool.InvokeDynamicEntry)invokeDynamicEntries.get(stringValue);
        if (invokeDynamicEntry == null) {
            invokeDynamicEntry = new ConstantPool.InvokeDynamicEntry(bootstrapMethodEntry, descriptorEntry);
            assert !(!invokeDynamicEntry.stringValue().equals(stringValue));
            invokeDynamicEntries.put(stringValue, invokeDynamicEntry);
        }
        return invokeDynamicEntry;
    }
    
    public static ConstantPool.BootstrapMethodEntry getBootstrapMethodEntry(ConstantPool.MethodHandleEntry methodHandleEntry, ConstantPool.Entry[] array) {
        Map<String, ConstantPool.BootstrapMethodEntry> bootstrapMethodEntries = Utils.getTLGlobals().getBootstrapMethodEntries();
        String stringValue = ConstantPool.BootstrapMethodEntry.stringValueOf(methodHandleEntry, array);
        ConstantPool.BootstrapMethodEntry bootstrapMethodEntry = (ConstantPool.BootstrapMethodEntry)bootstrapMethodEntries.get(stringValue);
        if (bootstrapMethodEntry == null) {
            bootstrapMethodEntry = new ConstantPool.BootstrapMethodEntry(methodHandleEntry, array);
            assert !(!bootstrapMethodEntry.stringValue().equals(stringValue));
            bootstrapMethodEntries.put(stringValue, bootstrapMethodEntry);
        }
        return bootstrapMethodEntry;
    }
    
    static boolean isMemberTag(byte b) {
        switch (b) {
            case 9:
            case 10:
            case 11: {
                return true;
            }
            default: {
                return false;
            }
        }
    }
    
    static byte numberTagOf(Number obj) {
        if ((obj instanceof Integer)) {
            return 3;
        }
        if ((obj instanceof Float)) {
            return 4;
        }
        if ((obj instanceof Long)) {
            return 5;
        }
        if ((obj instanceof Double)) {
            return 6;
        }
        throw new RuntimeException("bad literal value " + obj);
    }
    
    static boolean isRefKind(byte b) {
        return 1 <= b && b <= 9;
    }
    
    static String qualifiedStringValue(ConstantPool.Entry entry, ConstantPool.Entry entry2) {
        return qualifiedStringValue(entry.stringValue(), entry2.stringValue());
    }
    
    static String qualifiedStringValue(String str, String str2) {
        assert str.indexOf(".") < 0;
        return str + "." + str2;
    }
    
    static int compareSignatures(String s, String s2) {
        return compareSignatures(s, s2, null, (String[])null);
    }
    
    static int compareSignatures(String s, String anObject, String[] structureSignature, String[] structureSignature2) {
        int char1 = s.charAt(0);
        int char2 = anObject.charAt(0);
        if (char1 != '(' && char2 == '(') {
            return -1;
        }
        if (char2 != '(' && char1 == '(') {
            return 1;
        }
        if (structureSignature == null) {
            structureSignature = structureSignature(s);
        }
        if (structureSignature2 == null) {
            structureSignature2 = structureSignature(anObject);
        }
        if (structureSignature.length != structureSignature2.length) {
            return structureSignature.length - structureSignature2.length;
        }
        int length = structureSignature.length;
        while (--length >= 0) {
            int compareTo = structureSignature[length].compareTo(structureSignature2[length]);
            if (compareTo != 0) {
                return compareTo;
            }
            continue;
        }
        assert !(!s.equals(anObject));
        return 0;
    }
    
    static int countClassParts(ConstantPool.Utf8Entry utf8Entry) {
        int n = 0;
        String stringValue = utf8Entry.stringValue();
        for (int i = 0; i < stringValue.length(); ++i) {
            if (stringValue.charAt(i) == 'L') {
                ++n;
            }
        }
        return n;
    }
    
    static String flattenSignature(String[] array) {
        String s = array[0];
        if (array.length == 1) {
            return s;
        }
        int length = s.length();
        for (int i = 1; i < array.length; ++i) {
            length += array[i].length();
        }
        char[] array2 = new char[length];
        int dstBegin = 0;
        int n = 1;
        for (int j = 0; j < s.length(); ++j) {
            if ((array2[dstBegin++] = s.charAt(j)) == 'L') {
                String s2 = array[n++];
                s2.getChars(0, s2.length(), array2, dstBegin);
                dstBegin += s2.length();
            }
        }
        assert dstBegin == length;
        assert n == array.length;
        return new String(array2);
    }
    
    private static int skipTo(char ch, String s, int index) {
        index = s.indexOf((int)ch, index);
        return (index >= 0) ? index : s.length();
    }
    
    static String[] structureSignature(String s) {
        int index = s.indexOf(76);
        if (index < 0) {
            return new String[] { s };
        }
        char[] value = null;
        String[] array = null;
        int i = 0;
        while (i <= 1) {
            int n = 0;
            int n2 = 1;
            int skipTo = 0;
            int skipTo2 = 0;
            int n3 = 0;
            for (int j = index + 1; j > 0; n += j - n3, ++n2, n3 = skipTo2, j = s.indexOf(76, skipTo2) + 1) {
                if (skipTo < j) {
                    skipTo = skipTo(';', s, j);
                }
                if (skipTo2 < j) {
                    skipTo2 = skipTo('<', s, j);
                }
                skipTo2 = ((skipTo < skipTo2) ? skipTo : skipTo2);
                if (i != 0) {
                    s.getChars(n3, j, value, n);
                    array[n2] = s.substring(j, skipTo2);
                }
            }
            if (i != 0) {
                s.getChars(n3, s.length(), value, n);
                break;
            }
            value = new char[n + (s.length() - n3)];
            array = new String[n2];
            ++i;
        }
        array[0] = new String(value);
        return array;
    }
    
    public static ConstantPool.Index makeIndex(String s, ConstantPool.Entry[] array) {
        return new ConstantPool.Index(s, array);
    }
    
    public static ConstantPool.Index makeIndex(String s, Collection<ConstantPool.Entry> collection) {
        return new ConstantPool.Index(s, collection);
    }
    
    public static void sort(ConstantPool.Index index) {
        index.clearIndex();
        Arrays.sort(index.cpMap);
        if (verbose() > 2) {
            System.out.println("sorted " + index.dumpString());
        }
    }
    
    public static ConstantPool.Index[] partition(ConstantPool.Index index, int[] array) {
        ArrayList<Object> list = new ArrayList();
        ConstantPool.Entry[] cpMap = index.cpMap;
        assert array.length == cpMap.length;
        for (int i = 0; i < array.length; ++i) {
            int j = array[i];
            if (j >= 0) {
                while (j >= list.size()) {
                    list.add(null);
                }
                List<Object> list2 = (List)list.get(j);
                if (list2 == null) {
                    list.set(j, list2 = new ArrayList());
                }
                list2.add(cpMap[i]);
            }
        }
        ConstantPool.Index[] array2 = new ConstantPool.Index[list.size()];
        for (int k = 0; k < array2.length; ++k) {
            List<Object> list3 = (List)list.get(k);
            if (list3 != null) {
                array2[k] = new ConstantPool.Index(index.debugName + "/part#" + k, list3);
                assert array2[k].indexOf((ConstantPool.Entry)list3.get(0)) == 0;
            }
        }
        return array2;
    }
    
    public static ConstantPool.Index[] partitionByTag(ConstantPool.Index index) {
        ConstantPool.Entry[] cpMap = index.cpMap;
        int[] array = new int[cpMap.length];
        for (int i = 0; i < array.length; ++i) {
            ConstantPool.Entry entry = cpMap[i];
            array[i] = ((entry == null) ? -1 : entry.tag);
        }
        ConstantPool.Index[] partition = partition(index, array);
        for (int j = 0; j < partition.length; ++j) {
            if (partition[j] != null) {
                partition[j].debugName = tagName(j);
            }
        }
        if (partition.length < 19) {
            ConstantPool.Index[] array2 = new ConstantPool.Index[19];
            System.arraycopy(partition, 0, array2, 0, partition.length);
            partition = array2;
            return partition;
        }
        return partition;
    }
    
    public static void completeReferencesIn(Set<ConstantPool.Entry> set, boolean b) {
        completeReferencesIn(set, b, null);
    }
    
    public static void completeReferencesIn(Set<ConstantPool.Entry> c, boolean b, List<ConstantPool.BootstrapMethodEntry> list) {
        c.remove(null);
        ListIterator<ConstantPool.Entry> listIterator = new ArrayList(c).listIterator(c.size());
        while (listIterator.hasPrevious()) {
            ConstantPool.Entry entry = (ConstantPool.Entry)listIterator.previous();
            listIterator.remove();
            assert entry != null;
            if (b && entry.tag == 13) {
                ConstantPool.SignatureEntry signatureEntry = (ConstantPool.SignatureEntry)entry;
                ConstantPool.Utf8Entry utf8Entry = signatureEntry.asUtf8Entry();
                c.remove(signatureEntry);
                c.add(utf8Entry);
                entry = utf8Entry;
            }
            if (list != null && entry.tag == 17) {
                ConstantPool.BootstrapMethodEntry bootstrapMethodEntry = (ConstantPool.BootstrapMethodEntry)entry;
                c.remove(bootstrapMethodEntry);
                if (!list.contains(bootstrapMethodEntry)) {
                    list.add(bootstrapMethodEntry);
                }
            }
            int n = 0;
            while (true) {
                ConstantPool.Entry ref = entry.getRef(n);
                if (ref == null) {
                    break;
                }
                if (c.add(ref)) {
                    listIterator.add(ref);
                }
                ++n;
            }
        }
    }
    
    static double percent(int n, int n2) {
        return (double)(int)(10000.0 * (double)n / (double)n2 + 0.5) / 100.0;
    }
    
    public static String tagName(int i) {
        switch (i) {
            case 1: {
                return "Utf8";
            }
            case 3: {
                return "Integer";
            }
            case 4: {
                return "Float";
            }
            case 5: {
                return "Long";
            }
            case 6: {
                return "Double";
            }
            case 7: {
                return "Class";
            }
            case 8: {
                return "String";
            }
            case 9: {
                return "Fieldref";
            }
            case 10: {
                return "Methodref";
            }
            case 11: {
                return "InterfaceMethodref";
            }
            case 12: {
                return "NameandType";
            }
            case 15: {
                return "MethodHandle";
            }
            case 16: {
                return "MethodType";
            }
            case 18: {
                return "InvokeDynamic";
            }
            case 50: {
                return "**All";
            }
            case 0: {
                return "**None";
            }
            case 51: {
                return "**LoadableValue";
            }
            case 52: {
                return "**AnyMember";
            }
            case 53: {
                return "*FieldSpecific";
            }
            case 13: {
                return "*Signature";
            }
            case 17: {
                return "*BootstrapMethod";
            }
            default: {
                return "tag#" + i;
            }
        }
    }
    
    public static String refKindName(int i) {
        switch (i) {
            case 1: {
                return "getField";
            }
            case 2: {
                return "getStatic";
            }
            case 3: {
                return "putField";
            }
            case 4: {
                return "putStatic";
            }
            case 5: {
                return "invokeVirtual";
            }
            case 6: {
                return "invokeStatic";
            }
            case 7: {
                return "invokeSpecial";
            }
            case 8: {
                return "newInvokeSpecial";
            }
            case 9: {
                return "invokeInterface";
            }
            default: {
                return "refKind#" + i;
            }
        }
    }
    
    private static boolean verifyTagOrder(byte[] a) {
        int n = -1;
        int length = a.length;
        int i = 0;
        while (i < length) {
            int j = a[i];
            int k = ConstantPool.TAG_ORDER[j];
            assert k > 0 : "tag not found: " + j;
            assert ConstantPool.TAGS_IN_ORDER[k - 1] == j : "tag repeated: " + j + " => " + k + " => " + (int)ConstantPool.TAGS_IN_ORDER[k - 1];
            assert n < k : "tags not in order: " + Arrays.toString(a) + " at " + j;
            n = k;
            ++i;
        }
        return true;
    }
    
    static {
        noRefs = new ConstantPool.Entry[0];
        noClassRefs = new ConstantPool.ClassEntry[0];
        TAGS_IN_ORDER = new byte[] { 1, 3, 4, 5, 6, 8, 7, 13, 12, 9, 10, 11, 15, 16, 17, 18 };
        TAG_ORDER = new byte[19];
        for (int i = 0; i < ConstantPool.TAGS_IN_ORDER.length; ++i) {
            ConstantPool.TAG_ORDER[ConstantPool.TAGS_IN_ORDER[i]] = (byte)(i + 1);
        }
        NUMBER_TAGS = new byte[] { 3, 4, 5, 6 };
        EXTRA_TAGS = new byte[] { 15, 16, 17, 18 };
        LOADABLE_VALUE_TAGS = new byte[] { 3, 4, 5, 6, 8, 7, 15, 16 };
        ANY_MEMBER_TAGS = new byte[] { 9, 10, 11 };
        FIELD_SPECIFIC_TAGS = new byte[] { 3, 4, 5, 6, 8 };
        assert verifyTagOrder(ConstantPool.TAGS_IN_ORDER) && verifyTagOrder(ConstantPool.NUMBER_TAGS) && verifyTagOrder(ConstantPool.EXTRA_TAGS) && verifyTagOrder(ConstantPool.LOADABLE_VALUE_TAGS) && verifyTagOrder(ConstantPool.ANY_MEMBER_TAGS) && verifyTagOrder(ConstantPool.FIELD_SPECIFIC_TAGS);
        return;
    }
    
    public abstract static class Entry implements Comparable<Object>
    {
        protected final byte tag;
        protected int valueHash;
        
        protected Entry(byte tag) {
            this.tag = tag;
        }
        
        public final byte getTag() {
            return this.tag;
        }
        
        public final boolean tagEquals(int n) {
            return this.getTag() == n;
        }
        
        public ConstantPool.Entry getRef(int n) {
            return null;
        }
        
        public boolean eq(ConstantPool.Entry entry) {
            assert entry != null;
            return this == entry || this.equals(entry);
        }
        
        @Override
        public abstract boolean equals(Object p0);
        
        @Override
        public final int hashCode() {
            if (this.valueHash == 0) {
                this.valueHash = this.computeValueHash();
                if (this.valueHash == 0) {
                    this.valueHash = 1;
                }
            }
            return this.valueHash;
        }
        
        protected abstract int computeValueHash();
        
        @Override
        public abstract int compareTo(Object p0);
        
        protected int superCompareTo(Object o) {
            ConstantPool.Entry entry = (ConstantPool.Entry)o;
            if (this.tag != entry.tag) {
                return ConstantPool.TAG_ORDER[this.tag] - ConstantPool.TAG_ORDER[entry.tag];
            }
            return 0;
        }
        
        public final boolean isDoubleWord() {
            return this.tag == 6 || this.tag == 5;
        }
        
        public final boolean tagMatches(int n) {
            if (this.tag == n) {
                return true;
            }
            byte[] array;
            switch (n) {
                case 50: {
                    return true;
                }
                case 13: {
                    return this.tag == 1;
                }
                case 51: {
                    array = ConstantPool.LOADABLE_VALUE_TAGS;
                    break;
                }
                case 52: {
                    array = ConstantPool.ANY_MEMBER_TAGS;
                    break;
                }
                case 53: {
                    array = ConstantPool.FIELD_SPECIFIC_TAGS;
                    break;
                }
                default: {
                    return false;
                }
            }
            int length = array.length;
            int i = 0;
            while (i < length) {
                if (array[i] == this.tag) {
                    return true;
                }
                ++i;
            }
            return false;
        }
        
        @Override
        public String toString() {
            String str = this.stringValue();
            if (ConstantPool.verbose() > 4) {
                if (this.valueHash != 0) {
                    str = str + " hash=" + this.valueHash;
                }
                str = str + " id=" + System.identityHashCode(this);
            }
            return ConstantPool.tagName((int)this.tag) + "=" + str;
        }
        
        public abstract String stringValue();
    }
    
    public static class Utf8Entry extends ConstantPool.Entry
    {
        final String value;
        
        Utf8Entry(String s) {
            super((byte)1);
            this.value = s.intern();
            this.hashCode();
        }
        
        @Override
        protected int computeValueHash() {
            return this.value.hashCode();
        }
        
        @Override
        public boolean equals(Object o) {
            return o != null && o.getClass() == ConstantPool.Utf8Entry.class && ((ConstantPool.Utf8Entry)o).value.equals(this.value);
        }
        
        @Override
        public int compareTo(Object o) {
            int n = this.superCompareTo(o);
            if (n == 0) {
                n = this.value.compareTo(((ConstantPool.Utf8Entry)o).value);
            }
            return n;
        }
        
        @Override
        public String stringValue() {
            return this.value;
        }
    }
    
    public abstract static class LiteralEntry extends ConstantPool.Entry
    {
        protected LiteralEntry(byte b) {
            super(b);
        }
        
        public abstract Comparable<?> literalValue();
    }
    
    public static class NumberEntry extends ConstantPool.LiteralEntry
    {
        final Number value;
        
        NumberEntry(Number value) {
            super(ConstantPool.numberTagOf(value));
            this.value = value;
            this.hashCode();
        }
        
        @Override
        protected int computeValueHash() {
            return this.value.hashCode();
        }
        
        @Override
        public boolean equals(Object o) {
            return o != null && o.getClass() == ConstantPool.NumberEntry.class && ((ConstantPool.NumberEntry)o).value.equals(this.value);
        }
        
        @Override
        public int compareTo(Object o) {
            int n = this.superCompareTo(o);
            if (n == 0) {
                n = ((Comparable)this.value).compareTo(((ConstantPool.NumberEntry)o).value);
            }
            return n;
        }
        
        public Number numberValue() {
            return this.value;
        }
        
        @Override
        public Comparable<?> literalValue() {
            return (Comparable)this.value;
        }
        
        @Override
        public String stringValue() {
            return this.value.toString();
        }
    }
    
    public static class StringEntry extends ConstantPool.LiteralEntry
    {
        final ConstantPool.Utf8Entry ref;
        
        @Override
        public ConstantPool.Entry getRef(int n) {
            return (n == 0) ? this.ref : null;
        }
        
        StringEntry(ConstantPool.Entry entry) {
            super((byte)8);
            this.ref = (ConstantPool.Utf8Entry)entry;
            this.hashCode();
        }
        
        @Override
        protected int computeValueHash() {
            return this.ref.hashCode() + this.tag;
        }
        
        @Override
        public boolean equals(Object o) {
            return o != null && o.getClass() == ConstantPool.StringEntry.class && ((ConstantPool.StringEntry)o).ref.eq(this.ref);
        }
        
        @Override
        public int compareTo(Object o) {
            int n = this.superCompareTo(o);
            if (n == 0) {
                n = this.ref.compareTo(((ConstantPool.StringEntry)o).ref);
            }
            return n;
        }
        
        @Override
        public Comparable<?> literalValue() {
            return this.ref.stringValue();
        }
        
        @Override
        public String stringValue() {
            return this.ref.stringValue();
        }
    }
    
    public static class ClassEntry extends ConstantPool.Entry
    {
        final ConstantPool.Utf8Entry ref;
        
        @Override
        public ConstantPool.Entry getRef(int n) {
            return (n == 0) ? this.ref : null;
        }
        
        @Override
        protected int computeValueHash() {
            return this.ref.hashCode() + this.tag;
        }
        
        ClassEntry(ConstantPool.Entry entry) {
            super((byte)7);
            this.ref = (ConstantPool.Utf8Entry)entry;
            this.hashCode();
        }
        
        @Override
        public boolean equals(Object o) {
            return o != null && o.getClass() == ConstantPool.ClassEntry.class && ((ConstantPool.ClassEntry)o).ref.eq(this.ref);
        }
        
        @Override
        public int compareTo(Object o) {
            int n = this.superCompareTo(o);
            if (n == 0) {
                n = this.ref.compareTo(((ConstantPool.ClassEntry)o).ref);
            }
            return n;
        }
        
        @Override
        public String stringValue() {
            return this.ref.stringValue();
        }
    }
    
    public static class DescriptorEntry extends ConstantPool.Entry
    {
        final ConstantPool.Utf8Entry nameRef;
        final ConstantPool.SignatureEntry typeRef;
        
        @Override
        public ConstantPool.Entry getRef(int n) {
            if (n == 0) {
                return this.nameRef;
            }
            if (n == 1) {
                return this.typeRef;
            }
            return null;
        }
        
        DescriptorEntry(ConstantPool.Entry entry, ConstantPool.Entry signatureEntry) {
            super((byte)12);
            if ((signatureEntry instanceof ConstantPool.Utf8Entry)) {
                signatureEntry = ConstantPool.getSignatureEntry(signatureEntry.stringValue());
            }
            this.nameRef = (ConstantPool.Utf8Entry)entry;
            this.typeRef = (ConstantPool.SignatureEntry)signatureEntry;
            this.hashCode();
        }
        
        @Override
        protected int computeValueHash() {
            int hashCode = this.typeRef.hashCode();
            return this.nameRef.hashCode() + (hashCode << 8) ^ hashCode;
        }
        
        @Override
        public boolean equals(Object o) {
            if (o == null || o.getClass() != ConstantPool.DescriptorEntry.class) {
                return false;
            }
            ConstantPool.DescriptorEntry descriptorEntry = (ConstantPool.DescriptorEntry)o;
            return this.nameRef.eq(descriptorEntry.nameRef) && this.typeRef.eq(descriptorEntry.typeRef);
        }
        
        @Override
        public int compareTo(Object o) {
            int n = this.superCompareTo(o);
            if (n == 0) {
                ConstantPool.DescriptorEntry descriptorEntry = (ConstantPool.DescriptorEntry)o;
                n = this.typeRef.compareTo(descriptorEntry.typeRef);
                if (n == 0) {
                    n = this.nameRef.compareTo(descriptorEntry.nameRef);
                }
            }
            return n;
        }
        
        @Override
        public String stringValue() {
            return stringValueOf(this.nameRef, this.typeRef);
        }
        
        static String stringValueOf(ConstantPool.Entry entry, ConstantPool.Entry entry2) {
            return ConstantPool.qualifiedStringValue(entry2, entry);
        }
        
        public String prettyString() {
            return this.nameRef.stringValue() + this.typeRef.prettyString();
        }
        
        public boolean isMethod() {
            return this.typeRef.isMethod();
        }
        
        public byte getLiteralTag() {
            return this.typeRef.getLiteralTag();
        }
    }
    
    public static class MemberEntry extends ConstantPool.Entry
    {
        final ConstantPool.ClassEntry classRef;
        final ConstantPool.DescriptorEntry descRef;
        
        @Override
        public ConstantPool.Entry getRef(int n) {
            if (n == 0) {
                return this.classRef;
            }
            if (n == 1) {
                return this.descRef;
            }
            return null;
        }
        
        @Override
        protected int computeValueHash() {
            int hashCode = this.descRef.hashCode();
            return this.classRef.hashCode() + (hashCode << 8) ^ hashCode;
        }
        
        MemberEntry(byte b, ConstantPool.ClassEntry classRef, ConstantPool.DescriptorEntry descRef) {
            super(b);
            assert !(!ConstantPool.isMemberTag(b));
            this.classRef = classRef;
            this.descRef = descRef;
            this.hashCode();
            return;
        }
        
        @Override
        public boolean equals(Object o) {
            if (o == null || o.getClass() != ConstantPool.MemberEntry.class) {
                return false;
            }
            ConstantPool.MemberEntry memberEntry = (ConstantPool.MemberEntry)o;
            return this.classRef.eq(memberEntry.classRef) && this.descRef.eq(memberEntry.descRef);
        }
        
        @Override
        public int compareTo(Object o) {
            int n = this.superCompareTo(o);
            if (n == 0) {
                ConstantPool.MemberEntry memberEntry = (ConstantPool.MemberEntry)o;
                if (Utils.SORT_MEMBERS_DESCR_MAJOR) {
                    n = this.descRef.compareTo(memberEntry.descRef);
                }
                if (n == 0) {
                    n = this.classRef.compareTo(memberEntry.classRef);
                }
                if (n == 0) {
                    n = this.descRef.compareTo(memberEntry.descRef);
                }
            }
            return n;
        }
        
        @Override
        public String stringValue() {
            return stringValueOf(this.tag, this.classRef, this.descRef);
        }
        
        static String stringValueOf(byte i, ConstantPool.ClassEntry classEntry, ConstantPool.DescriptorEntry descriptorEntry) {
            assert !(!ConstantPool.isMemberTag(i));
            String string;
            switch (i) {
                case 9: {
                    string = "Field:";
                    break;
                }
                case 10: {
                    string = "Method:";
                    break;
                }
                case 11: {
                    string = "IMethod:";
                    break;
                }
                default: {
                    string = (int)i + "???";
                    break;
                }
            }
            return string + ConstantPool.qualifiedStringValue(classEntry, descriptorEntry);
        }
        
        public boolean isMethod() {
            return this.descRef.isMethod();
        }
    }
    
    public static class SignatureEntry extends ConstantPool.Entry
    {
        final ConstantPool.Utf8Entry formRef;
        final ConstantPool.ClassEntry[] classRefs;
        String value;
        ConstantPool.Utf8Entry asUtf8Entry;
        
        @Override
        public ConstantPool.Entry getRef(int n) {
            if (n == 0) {
                return this.formRef;
            }
            return (n - 1 < this.classRefs.length) ? this.classRefs[n - 1] : null;
        }
        
        SignatureEntry(String intern) {
            super((byte)13);
            intern = intern.intern();
            this.value = intern;
            String[] structureSignature = ConstantPool.structureSignature(intern);
            this.formRef = ConstantPool.getUtf8Entry(structureSignature[0]);
            this.classRefs = new ConstantPool.ClassEntry[structureSignature.length - 1];
            for (int i = 1; i < structureSignature.length; ++i) {
                this.classRefs[i - 1] = ConstantPool.getClassEntry(structureSignature[i]);
            }
            this.hashCode();
        }
        
        @Override
        protected int computeValueHash() {
            this.stringValue();
            return this.value.hashCode() + this.tag;
        }
        
        public ConstantPool.Utf8Entry asUtf8Entry() {
            if (this.asUtf8Entry == null) {
                this.asUtf8Entry = ConstantPool.getUtf8Entry(this.stringValue());
            }
            return this.asUtf8Entry;
        }
        
        @Override
        public boolean equals(Object o) {
            return o != null && o.getClass() == ConstantPool.SignatureEntry.class && ((ConstantPool.SignatureEntry)o).value.equals(this.value);
        }
        
        @Override
        public int compareTo(Object o) {
            int n = this.superCompareTo(o);
            if (n == 0) {
                n = ConstantPool.compareSignatures(this.value, ((ConstantPool.SignatureEntry)o).value);
            }
            return n;
        }
        
        @Override
        public String stringValue() {
            if (this.value == null) {
                this.value = stringValueOf(this.formRef, this.classRefs);
            }
            return this.value;
        }
        
        static String stringValueOf(ConstantPool.Utf8Entry utf8Entry, ConstantPool.ClassEntry[] array) {
            String[] array2 = new String[1 + array.length];
            array2[0] = utf8Entry.stringValue();
            for (int i = 1; i < array2.length; ++i) {
                array2[i] = array[i - 1].stringValue();
            }
            return ConstantPool.flattenSignature(array2).intern();
        }
        
        public int computeSize(boolean b) {
            String stringValue = this.formRef.stringValue();
            int i = 0;
            int index = 1;
            if (this.isMethod()) {
                i = 1;
                index = stringValue.indexOf(41);
            }
            int n = 0;
            for (; i < index; ++i) {
                switch (stringValue.charAt(i)) {
                    case 'D':
                    case 'J': {
                        if (b) {
                            ++n;
                            break;
                        }
                        break;
                    }
                    case '[': {
                        while (stringValue.charAt(i) == '[') {
                            ++i;
                        }
                        break;
                    }
                    case ';': {
                        continue;
                    }
                    default: {
                        assert 0 <= "BSCIJFDZLV([".indexOf((int)stringValue.charAt(i));
                        break;
                    }
                }
                ++n;
            }
            return n;
        }
        
        public boolean isMethod() {
            return this.formRef.stringValue().charAt(0) == '(';
        }
        
        public byte getLiteralTag() {
            switch (this.formRef.stringValue().charAt(0)) {
                case 'I': {
                    return 3;
                }
                case 'J': {
                    return 5;
                }
                case 'F': {
                    return 4;
                }
                case 'D': {
                    return 6;
                }
                case 'B':
                case 'C':
                case 'S':
                case 'Z': {
                    return 3;
                }
                case 'L': {
                    return 8;
                }
                default: {
                    assert false;
                    return 0;
                }
            }
        }
        
        public String prettyString() {
            String s;
            if (this.isMethod()) {
                String stringValue = this.formRef.stringValue();
                s = stringValue.substring(0, 1 + stringValue.indexOf(41));
            }
            else {
                s = "/" + this.formRef.stringValue();
            }
            int index;
            while ((index = s.indexOf(59)) >= 0) {
                s = s.substring(0, index) + s.substring(index + 1);
            }
            return s;
        }
    }
    
    public static class MethodHandleEntry extends ConstantPool.Entry
    {
        final int refKind;
        final ConstantPool.MemberEntry memRef;
        
        @Override
        public ConstantPool.Entry getRef(int n) {
            return (n == 0) ? this.memRef : null;
        }
        
        @Override
        protected int computeValueHash() {
            int refKind = this.refKind;
            return this.memRef.hashCode() + (refKind << 8) ^ refKind;
        }
        
        MethodHandleEntry(byte refKind, ConstantPool.MemberEntry memRef) {
            super((byte)15);
            assert !(!ConstantPool.isRefKind(refKind));
            this.refKind = refKind;
            this.memRef = memRef;
            this.hashCode();
            return;
        }
        
        @Override
        public boolean equals(Object o) {
            if (o == null || o.getClass() != ConstantPool.MethodHandleEntry.class) {
                return false;
            }
            ConstantPool.MethodHandleEntry methodHandleEntry = (ConstantPool.MethodHandleEntry)o;
            return this.refKind == methodHandleEntry.refKind && this.memRef.eq(methodHandleEntry.memRef);
        }
        
        @Override
        public int compareTo(Object o) {
            int n = this.superCompareTo(o);
            if (n == 0) {
                ConstantPool.MethodHandleEntry methodHandleEntry = (ConstantPool.MethodHandleEntry)o;
                if (Utils.SORT_HANDLES_KIND_MAJOR) {
                    n = this.refKind - methodHandleEntry.refKind;
                }
                if (n == 0) {
                    n = this.memRef.compareTo(methodHandleEntry.memRef);
                }
                if (n == 0) {
                    n = this.refKind - methodHandleEntry.refKind;
                }
            }
            return n;
        }
        
        public static String stringValueOf(int n, ConstantPool.MemberEntry memberEntry) {
            return ConstantPool.refKindName(n) + ":" + memberEntry.stringValue();
        }
        
        @Override
        public String stringValue() {
            return stringValueOf(this.refKind, this.memRef);
        }
    }
    
    public static class MethodTypeEntry extends ConstantPool.Entry
    {
        final ConstantPool.SignatureEntry typeRef;
        
        @Override
        public ConstantPool.Entry getRef(int n) {
            return (n == 0) ? this.typeRef : null;
        }
        
        @Override
        protected int computeValueHash() {
            return this.typeRef.hashCode() + this.tag;
        }
        
        MethodTypeEntry(ConstantPool.SignatureEntry typeRef) {
            super((byte)16);
            this.typeRef = typeRef;
            this.hashCode();
        }
        
        @Override
        public boolean equals(Object o) {
            return o != null && o.getClass() == ConstantPool.MethodTypeEntry.class && this.typeRef.eq(((ConstantPool.MethodTypeEntry)o).typeRef);
        }
        
        @Override
        public int compareTo(Object o) {
            int n = this.superCompareTo(o);
            if (n == 0) {
                n = this.typeRef.compareTo(((ConstantPool.MethodTypeEntry)o).typeRef);
            }
            return n;
        }
        
        @Override
        public String stringValue() {
            return this.typeRef.stringValue();
        }
    }
    
    public static class InvokeDynamicEntry extends ConstantPool.Entry
    {
        final ConstantPool.BootstrapMethodEntry bssRef;
        final ConstantPool.DescriptorEntry descRef;
        
        @Override
        public ConstantPool.Entry getRef(int n) {
            if (n == 0) {
                return this.bssRef;
            }
            if (n == 1) {
                return this.descRef;
            }
            return null;
        }
        
        @Override
        protected int computeValueHash() {
            int hashCode = this.descRef.hashCode();
            return this.bssRef.hashCode() + (hashCode << 8) ^ hashCode;
        }
        
        InvokeDynamicEntry(ConstantPool.BootstrapMethodEntry bssRef, ConstantPool.DescriptorEntry descRef) {
            super((byte)18);
            this.bssRef = bssRef;
            this.descRef = descRef;
            this.hashCode();
        }
        
        @Override
        public boolean equals(Object o) {
            if (o == null || o.getClass() != ConstantPool.InvokeDynamicEntry.class) {
                return false;
            }
            ConstantPool.InvokeDynamicEntry invokeDynamicEntry = (ConstantPool.InvokeDynamicEntry)o;
            return this.bssRef.eq(invokeDynamicEntry.bssRef) && this.descRef.eq(invokeDynamicEntry.descRef);
        }
        
        @Override
        public int compareTo(Object o) {
            int n = this.superCompareTo(o);
            if (n == 0) {
                ConstantPool.InvokeDynamicEntry invokeDynamicEntry = (ConstantPool.InvokeDynamicEntry)o;
                if (Utils.SORT_INDY_BSS_MAJOR) {
                    n = this.bssRef.compareTo(invokeDynamicEntry.bssRef);
                }
                if (n == 0) {
                    n = this.descRef.compareTo(invokeDynamicEntry.descRef);
                }
                if (n == 0) {
                    n = this.bssRef.compareTo(invokeDynamicEntry.bssRef);
                }
            }
            return n;
        }
        
        @Override
        public String stringValue() {
            return stringValueOf(this.bssRef, this.descRef);
        }
        
        static String stringValueOf(ConstantPool.BootstrapMethodEntry bootstrapMethodEntry, ConstantPool.DescriptorEntry descriptorEntry) {
            return "Indy:" + bootstrapMethodEntry.stringValue() + "." + descriptorEntry.stringValue();
        }
    }
    
    public static class BootstrapMethodEntry extends ConstantPool.Entry
    {
        final ConstantPool.MethodHandleEntry bsmRef;
        final ConstantPool.Entry[] argRefs;
        
        @Override
        public ConstantPool.Entry getRef(int n) {
            if (n == 0) {
                return this.bsmRef;
            }
            if (n - 1 < this.argRefs.length) {
                return this.argRefs[n - 1];
            }
            return null;
        }
        
        @Override
        protected int computeValueHash() {
            int hashCode = this.bsmRef.hashCode();
            return Arrays.hashCode(this.argRefs) + (hashCode << 8) ^ hashCode;
        }
        
        BootstrapMethodEntry(ConstantPool.MethodHandleEntry bsmRef, ConstantPool.Entry[] array) {
            super((byte)17);
            this.bsmRef = bsmRef;
            this.argRefs = (ConstantPool.Entry[])array.clone();
            this.hashCode();
        }
        
        @Override
        public boolean equals(Object o) {
            if (o == null || o.getClass() != ConstantPool.BootstrapMethodEntry.class) {
                return false;
            }
            ConstantPool.BootstrapMethodEntry bootstrapMethodEntry = (ConstantPool.BootstrapMethodEntry)o;
            return this.bsmRef.eq(bootstrapMethodEntry.bsmRef) && Arrays.equals(this.argRefs, bootstrapMethodEntry.argRefs);
        }
        
        @Override
        public int compareTo(Object o) {
            int n = this.superCompareTo(o);
            if (n == 0) {
                ConstantPool.BootstrapMethodEntry bootstrapMethodEntry = (ConstantPool.BootstrapMethodEntry)o;
                if (Utils.SORT_BSS_BSM_MAJOR) {
                    n = this.bsmRef.compareTo(bootstrapMethodEntry.bsmRef);
                }
                if (n == 0) {
                    n = compareArgArrays(this.argRefs, bootstrapMethodEntry.argRefs);
                }
                if (n == 0) {
                    n = this.bsmRef.compareTo(bootstrapMethodEntry.bsmRef);
                }
            }
            return n;
        }
        
        @Override
        public String stringValue() {
            return stringValueOf(this.bsmRef, this.argRefs);
        }
        
        static String stringValueOf(ConstantPool.MethodHandleEntry methodHandleEntry, ConstantPool.Entry[] array) {
            StringBuffer sb = new StringBuffer(methodHandleEntry.stringValue());
            int n = '<';
            for (int length = array.length, i = 0; i < length; n = ';', ++i) {
                sb.append((char)n).append(array[i].stringValue());
            }
            if (n == '<') {
                sb.append((char)n);
            }
            sb.append('>');
            return sb.toString();
        }
        
        static int compareArgArrays(ConstantPool.Entry[] array, ConstantPool.Entry[] array2) {
            int compareTo = array.length - array2.length;
            if (compareTo != 0) {
                return compareTo;
            }
            int i = 0;
            while (i < array.length) {
                compareTo = array[i].compareTo(array2[i]);
                if (compareTo != 0) {
                    break;
                }
                ++i;
            }
            return compareTo;
        }
    }
    
    public static final class Index extends AbstractList<ConstantPool.Entry>
    {
        protected String debugName;
        protected ConstantPool.Entry[] cpMap;
        protected boolean flattenSigs;
        protected ConstantPool.Entry[] indexKey;
        protected int[] indexValue;
        
        protected ConstantPool.Entry[] getMap() {
            return this.cpMap;
        }
        
        protected Index(String debugName) {
            this.debugName = debugName;
        }
        
        protected Index(String s, ConstantPool.Entry[] map) {
            this(s);
            this.setMap(map);
        }
        
        protected void setMap(ConstantPool.Entry[] cpMap) {
            this.clearIndex();
            this.cpMap = cpMap;
        }
        
        protected Index(String s, Collection<ConstantPool.Entry> map) {
            this(s);
            this.setMap(map);
        }
        
        protected void setMap(Collection<ConstantPool.Entry> collection) {
            collection.toArray(this.cpMap = new ConstantPool.Entry[collection.size()]);
            this.setMap(this.cpMap);
        }
        
        @Override
        public int size() {
            return this.cpMap.length;
        }
        
        @Override
        public ConstantPool.Entry get(int n) {
            return this.cpMap[n];
        }
        
        public ConstantPool.Entry getEntry(int n) {
            return this.cpMap[n];
        }
        
        private int findIndexOf(ConstantPool.Entry entry) {
            if (this.indexKey == null) {
                this.initializeIndex();
            }
            int indexLocation = this.findIndexLocation(entry);
            if (this.indexKey[indexLocation] != entry) {
                if (this.flattenSigs && entry.tag == 13) {
                    return this.findIndexOf(((ConstantPool.SignatureEntry)entry).asUtf8Entry());
                }
                return -1;
            }
            else {
                int n = this.indexValue[indexLocation];
                assert !(!entry.equals(this.cpMap[n]));
                return n;
            }
        }
        
        public boolean contains(ConstantPool.Entry entry) {
            return this.findIndexOf(entry) >= 0;
        }
        
        public int indexOf(ConstantPool.Entry obj) {
            int index = this.findIndexOf(obj);
            if (index < 0 && ConstantPool.verbose() > 0) {
                System.out.println("not found: " + obj);
                System.out.println("       in: " + this.dumpString());
                Thread.dumpStack();
            }
            assert index >= 0;
            return index;
        }
        
        public int lastIndexOf(ConstantPool.Entry entry) {
            return this.indexOf(entry);
        }
        
        public boolean assertIsSorted() {
            int i = 1;
            while (i < this.cpMap.length) {
                if (this.cpMap[i - 1].compareTo(this.cpMap[i]) > 0) {
                    System.out.println("Not sorted at " + (i - 1) + "/" + i + ": " + this.dumpString());
                    return false;
                }
                ++i;
            }
            return true;
        }
        
        protected void clearIndex() {
            this.indexKey = null;
            this.indexValue = null;
        }
        
        private int findIndexLocation(ConstantPool.Entry entry) {
            int length = this.indexKey.length;
            int hashCode = entry.hashCode();
            int n = hashCode & length - 1;
            int n2 = (hashCode >>> 8 | 0x1) & length - 1;
            while (true) {
                ConstantPool.Entry entry2 = this.indexKey[n];
                if (entry2 == entry || entry2 == null) {
                    break;
                }
                n += n2;
                if (n >= length) {
                    n -= length;
                }
                else {
                    continue;
                }
            }
            return n;
        }
        
        private void initializeIndex() {
            if (ConstantPool.verbose() > 2) {
                System.out.println("initialize Index " + this.debugName + " [" + this.size() + "]");
            }
            int n;
            int i;
            for (n = (int)((double)(this.cpMap.length + 10) * 1.5), i = 1; i < n; i <<= 1) {}
            this.indexKey = new ConstantPool.Entry[i];
            this.indexValue = new int[i];
            for (int j = 0; j < this.cpMap.length; ++j) {
                ConstantPool.Entry entry = this.cpMap[j];
                if (entry != null) {
                    int indexLocation = this.findIndexLocation(entry);
                    assert this.indexKey[indexLocation] == null;
                    this.indexKey[indexLocation] = entry;
                    this.indexValue[indexLocation] = j;
                }
            }
        }
        
        public ConstantPool.Entry[] toArray(ConstantPool.Entry[] a) {
            int size = this.size();
            if (a.length < size) {
                return (ConstantPool.Entry[])super.toArray(a);
            }
            System.arraycopy(this.cpMap, 0, a, 0, size);
            if (a.length > size) {
                a[size] = null;
                return a;
            }
            return a;
        }
        
        @Override
        public ConstantPool.Entry[] toArray() {
            return this.toArray(new ConstantPool.Entry[this.size()]);
        }
        
        public Object clone() {
            return new ConstantPool.Index(this.debugName, (ConstantPool.Entry[])this.cpMap.clone());
        }
        
        @Override
        public String toString() {
            return "Index " + this.debugName + " [" + this.size() + "]";
        }
        
        public String dumpString() {
            String s = this.toString() + " {\n";
            for (int i = 0; i < this.cpMap.length; ++i) {
                s = s + "    " + i + ": " + this.cpMap[i] + "\n";
            }
            return s + "}";
        }
        
        static int access$000(ConstantPool.Index index, ConstantPool.Entry entry) {
            return index.findIndexOf(entry);
        }
    }
    
    public static class IndexGroup
    {
        private ConstantPool.Index[] indexByTag;
        private ConstantPool.Index[] indexByTagGroup;
        private int[] untypedFirstIndexByTag;
        private int totalSizeQQ;
        private ConstantPool.Index[][] indexByTagAndClass;
        static final /* synthetic */ boolean $assertionsDisabled;
        
        public IndexGroup() {
            this.indexByTag = new ConstantPool.Index[19];
        }
        
        private ConstantPool.Index makeTagGroupIndex(byte b, byte[] array) {
            if (this.indexByTagGroup == null) {
                this.indexByTagGroup = new ConstantPool.Index[4];
            }
            int n = b - 50;
            assert this.indexByTagGroup[n] == null;
            int n2 = 0;
            ConstantPool.Entry[] array2 = null;
            for (int i = 1; i <= 2; ++i) {
                this.untypedIndexOf(null);
                for (int n3 : array) {
                    ConstantPool.Index index = this.indexByTag[n3];
                    if (index != null) {
                        int length2 = index.cpMap.length;
                        if (length2 != 0) {
                            if (!ConstantPool.IndexGroup.$assertionsDisabled) {
                                if ((b != 50) ? (n2 < this.untypedFirstIndexByTag[n3]) : (n2 == this.untypedFirstIndexByTag[n3])) {}
                                throw new AssertionError();
                            }
                            if (array2 != null) {
                                assert array2[n2] == null;
                                assert array2[n2 + length2 - 1] == null;
                                System.arraycopy(index.cpMap, 0, array2, n2, length2);
                            }
                            n2 += length2;
                        }
                    }
                }
                if (array2 == null) {
                    assert i == 1;
                    array2 = new ConstantPool.Entry[n2];
                    n2 = 0;
                }
            }
            return this.indexByTagGroup[n] = new ConstantPool.Index(ConstantPool.tagName((int)b), array2);
        }
        
        public int untypedIndexOf(ConstantPool.Entry entry) {
            if (this.untypedFirstIndexByTag == null) {
                this.untypedFirstIndexByTag = new int[20];
                int n = 0;
                for (int i = 0; i < ConstantPool.TAGS_IN_ORDER.length; ++i) {
                    int n2 = ConstantPool.TAGS_IN_ORDER[i];
                    ConstantPool.Index index = this.indexByTag[n2];
                    if (index != null) {
                        int length = index.cpMap.length;
                        this.untypedFirstIndexByTag[n2] = n;
                        n += length;
                    }
                }
                this.untypedFirstIndexByTag[19] = n;
            }
            if (entry == null) {
                return -1;
            }
            int tag = entry.tag;
            ConstantPool.Index index2 = this.indexByTag[tag];
            if (index2 == null) {
                return -1;
            }
            int access$000 = ConstantPool.Index.access$000(index2, entry);
            if (access$000 >= 0) {
                access$000 += this.untypedFirstIndexByTag[tag];
            }
            return access$000;
        }
        
        public void initIndexByTag(byte b, ConstantPool.Index index) {
            assert this.indexByTag[b] == null;
            ConstantPool.Entry[] cpMap = index.cpMap;
            int i = 0;
            while (i < cpMap.length) {
                assert cpMap[i].tag == b;
                ++i;
            }
            if (b == 1 && !ConstantPool.IndexGroup.$assertionsDisabled && cpMap.length != 0 && !cpMap[0].stringValue().equals("")) {
                throw new AssertionError();
            }
            this.indexByTag[b] = index;
            this.untypedFirstIndexByTag = null;
            this.indexByTagGroup = null;
            if (this.indexByTagAndClass != null) {
                this.indexByTagAndClass[b] = null;
            }
            return;
        }
        
        public ConstantPool.Index getIndexByTag(byte b) {
            if (b >= 50) {
                return this.getIndexByTagGroup(b);
            }
            ConstantPool.Index index = this.indexByTag[b];
            if (index == null) {
                index = new ConstantPool.Index(ConstantPool.tagName((int)b), new ConstantPool.Entry[0]);
                this.indexByTag[b] = index;
                return index;
            }
            return index;
        }
        
        private ConstantPool.Index getIndexByTagGroup(byte i) {
            if (this.indexByTagGroup != null) {
                ConstantPool.Index index = this.indexByTagGroup[i - 50];
                if (index != null) {
                    return index;
                }
            }
            switch (i) {
                case 50: {
                    return this.makeTagGroupIndex((byte)50, ConstantPool.TAGS_IN_ORDER);
                }
                case 51: {
                    return this.makeTagGroupIndex((byte)51, ConstantPool.LOADABLE_VALUE_TAGS);
                }
                case 52: {
                    return this.makeTagGroupIndex((byte)52, ConstantPool.ANY_MEMBER_TAGS);
                }
                case 53: {
                    return null;
                }
                default: {
                    throw new AssertionError("bad tag group " + (int)i);
                }
            }
        }
        
        public ConstantPool.Index getMemberIndex(byte b, ConstantPool.ClassEntry classEntry) {
            if (classEntry == null) {
                throw new RuntimeException("missing class reference for " + ConstantPool.tagName((int)b));
            }
            if (this.indexByTagAndClass == null) {
                this.indexByTagAndClass = new ConstantPool.Index[19][];
            }
            ConstantPool.Index indexByTag = this.getIndexByTag((byte)7);
            ConstantPool.Index[] partition = this.indexByTagAndClass[b];
            if (partition == null) {
                ConstantPool.Index indexByTag2 = this.getIndexByTag(b);
                int[] array = new int[indexByTag2.size()];
                for (int i = 0; i < array.length; ++i) {
                    array[i] = indexByTag.indexOf(((ConstantPool.MemberEntry)indexByTag2.get(i)).classRef);
                }
                partition = ConstantPool.partition(indexByTag2, array);
                int j = 0;
                while (j < partition.length) {
                    assert !(!partition[j].assertIsSorted());
                    ++j;
                }
                this.indexByTagAndClass[b] = partition;
            }
            return partition[indexByTag.indexOf(classEntry)];
        }
        
        public int getOverloadingIndex(ConstantPool.MemberEntry memberEntry) {
            ConstantPool.Index memberIndex = this.getMemberIndex(memberEntry.tag, memberEntry.classRef);
            ConstantPool.Utf8Entry nameRef = memberEntry.descRef.nameRef;
            int n = 0;
            int i = 0;
            while (i < memberIndex.cpMap.length) {
                ConstantPool.MemberEntry memberEntry2 = (ConstantPool.MemberEntry)memberIndex.cpMap[i];
                if (memberEntry2.equals(memberEntry)) {
                    return n;
                }
                if (memberEntry2.descRef.nameRef.equals(nameRef)) {
                    ++n;
                }
                ++i;
            }
            throw new RuntimeException("should not reach here");
        }
        
        public ConstantPool.MemberEntry getOverloadingForIndex(byte b, ConstantPool.ClassEntry classEntry, String anObject, int n) {
            assert !(!anObject.equals(anObject.intern()));
            ConstantPool.Index memberIndex = this.getMemberIndex(b, classEntry);
            int n2 = 0;
            for (int i = 0; i < memberIndex.cpMap.length; ++i) {
                ConstantPool.MemberEntry memberEntry = (ConstantPool.MemberEntry)memberIndex.cpMap[i];
                if (memberEntry.descRef.nameRef.stringValue().equals(anObject)) {
                    if (n2 == n) {
                        return memberEntry;
                    }
                    ++n2;
                }
            }
            throw new RuntimeException("should not reach here");
        }
        
        public boolean haveNumbers() {
            byte[] number_TAGS = ConstantPool.NUMBER_TAGS;
            int length = number_TAGS.length;
            int i = 0;
            while (i < length) {
                if (this.getIndexByTag(number_TAGS[i]).size() > 0) {
                    return true;
                }
                ++i;
            }
            return false;
        }
        
        public boolean haveExtraTags() {
            byte[] extra_TAGS = ConstantPool.EXTRA_TAGS;
            int length = extra_TAGS.length;
            int i = 0;
            while (i < length) {
                if (this.getIndexByTag(extra_TAGS[i]).size() > 0) {
                    return true;
                }
                ++i;
            }
            return false;
        }
    }
}
